import pandas as pd
from sklearn.preprocessing import LabelEncoder, Imputer, StandardScaler
from collections import Counter

def drop_missing(df):
    df = df.dropna()
    return df

def one_hot_encode(df, categorical_cols = None):
    if categorical_cols:
        le = LabelEncoder()
        for feature in categorical_cols:
            try:
                df[feature] = le.fit_transform(df[feature])
            except:
                print("Encoding error encountered for %s" % feature)
    return df

def normalize(df, categorical_cols = None):
    numerical_cols = list(df.select_dtypes(include = ['float64', 'int64']))
    numerical_cols = list((Counter(numerical_cols) - Counter(categorical_cols)).elements())
    ss = StandardScaler()
    for feature in numerical_cols:
        try:
            df[feature] = ss.fit_transform(df[feature].reshape(-1,1))
        except:
            print("Normalization error encountered for %s" % feature)
    return df

def sample(df, proportion = 1.0):
    return df.sample(frac = proportion) if proportion != 1.0 else df



import matplotlib as mpl
mpl.use('Agg')  # to run matplotlib without a GUI

import matplotlib.pyplot as plt
from math import sqrt, log
import numpy as np
import pandas as pd
import os
from sklearn.cluster import KMeans

from .base_model import Base
from .preprocessor import drop_missing, one_hot_encode, normalize, sample


class Kmeans(Base):

    def __init__(self, df=pd.DataFrame(), categorical_cols=None, id_col=None):
        super().__init__(df, categorical_cols, id_col)
        self.original_df, self.processed_df = self._preprocess(
            self.original_df)
        self.gaps = pd.Series()
        self.gaps_with_error = pd.Series()
        self.best_k = None
        self.labels = pd.Series(name = 'labels')

    def fit(self, n_clusters = 8):
        kmeans = KMeans(n_clusters = n_clusters).fit(self.processed_df)
        self.labels = pd.Series(data = kmeans.labels_, name = 'labels')
        return kmeans

    def plot(self, save_loc = 'ezcluster_files/'):
        if not os.path.exists(save_loc):
            os.makedirs(save_loc)
        self._plot_elbow(self.gaps, save_loc)
        self._plot_diff(self.gaps_with_error, save_loc)

    def optimal_k(self, min_k=1, max_k=10, num_iters = 10):
        if not self.best_k:
            self.gaps, self.gaps_with_error = self._gap_statistic(
                self.processed_df, min_k, max_k, num_iters)
            if self.gaps_with_error.iloc[0] < 0:
                for i, val in self.gaps_with_error.iteritems():
                    if val > 0:
                        k = i
                        self.best_k = i
                        break
        return self.best_k

    def save_instance(self, filename = 'ezcluster_files/ezc.pkl'):
        if '/' in filename:
            folder, name = filename.split('/')
            if not os.path.exists(folder):
                os.makedirs(folder)
        super().save_instance(filename)

    def load_instance(self, filename = 'ezcluster_files/ezc.pkl'):
        ezc = super().load_instance(filename)
        return ezc

    def write_csv(self, filename = 'ezcluster_files/data.csv'):
        if '/' in filename:
            folder, name = filename.split('/')
            if not os.path.exists(folder):
                os.makedirs(folder)
        df = self.original_df.copy()
        df = df.join(self.labels)
        super().write_csv(df, filename)

    def _preprocess(self, df):
        df = drop_missing(df)
        df = sample(df)
        original = df.copy()
        processed = one_hot_encode(df, categorical_cols=self.categorical_cols)
        processed = normalize(df, categorical_cols=self.categorical_cols)
        return original, processed

    def _plot_elbow(self, gaps, save_loc):
        plt.figure()
        gaps.plot()
        plt.xlabel("Number of clusters K")
        plt.ylabel("Gap statistic")
        plt.savefig(os.path.join(save_loc, "gap_statistic.png"))
        plt.close()

    def _plot_diff(self, gaps_with_error, save_loc):
        plt.figure()
        gaps_with_error.plot(kind = 'bar')
        plt.xlabel("Number of clusters K")
        plt.ylabel("gap(k) - (gap(k+1) - err(k+1))")
        plt.savefig(os.path.join(save_loc, "gaps_with_error.png"))
        plt.close()

    def _gap_statistic(self, df, min_k, max_k, num_iters):
        k_range = range(min_k, max_k + 1)
        def get_rand_data(col):
            rng = col.max() - col.min()
            return pd.Series(np.random.random_sample(len(col))*rng + col.min())

        def iterate_k_means(df, n_clusters, num_iters):
            rng =  range(1, num_iters + 1)
            vals = pd.Series(index = rng)
            for i in rng:
                k = KMeans(n_clusters = n_clusters)
                k.fit(df)
                vals[i] = k.inertia_
            return vals

        gaps = pd.Series(index = k_range)
        simulation_error = pd.Series(index = k_range)
        for k in k_range:
            km = KMeans(n_clusters = k, n_init = 3)
            km.fit(df)
            ref = df.apply(get_rand_data)
            vals = iterate_k_means(ref, n_clusters = k, num_iters = num_iters)
            ref_intertia = vals.mean()
            gap = log(ref_intertia) - log(km.inertia_)
            gaps[k] = gap
            ref_std = vals.std()
            sd = sqrt(1 + (1/num_iters)) * ref_std
            simulation_error[k] = sd

        gaps_with_error = pd.Series(index = range(min_k, max_k))
        for i in range(min_k, max_k):
            gaps_with_error[i] = gaps[i] - (gaps[i+1] - simulation_error[i+1])
        return gaps, gaps_with_error



import pandas as pd
from sklearn.externals import joblib

class Base:
    def __init__(self, dataset, categorical_cols, id_col):
        self.categorical_cols = categorical_cols
        if id_col:
            self.ids = dataset[id_col]
            dataset.drop(id_col, axis = 1, inplace = True)
        self.original_df = dataset

    def save_instance(self, filename):
        joblib.dump(self, filename)

    def load_instance(self, filename):
        ezc = joblib.load(filename)
        return ezc

    def write_csv(self, df, filename):
        df.to_csv(filename, index = False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import os
import re
import sys
from subprocess import PIPE, run

try:
    import pypandoc
    readme = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
    readme = ''


package = 'ezcluster'
requirements = [

]
test_requirements = [

]


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("^__version__ = ['\"]([^'\"]+)['\"]",
                     init_py, re.MULTILINE).group(1)


def get_author(package):
    """
    Return package author as listed in `__author__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("^__author__ = ['\"]([^'\"]+)['\"]",
                     init_py, re.MULTILINE).group(1)


def get_email(package):
    """
    Return package email as listed in `__email__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("^__email__ = ['\"]([^'\"]+)['\"]",
                     init_py, re.MULTILINE).group(1)


# python setup.py register
if sys.argv[-1] == 'publish':
    os.system("python setup.py sdist upload")
    args = {'version': get_version(package)}
    print("You probably want to also tag the version now:")
    print("  git tag -a %(version)s -m 'version %(version)s'" % args)
    print("  git push --tags")
    sys.exit()

prev_tag = run("git tag", shell=True, stdout=PIPE, universal_newlines=True).stdout[:-1]
run("git tag -d " + prev_tag, shell = True)
run("git push --delete origin " + prev_tag, shell = True)
command = "git tag -a " + get_version(package) + " -m 'version " + get_version(package) + "'"
run(command, shell = True)
run("git push --tags", shell = True)

setup(
    name='ezcluster',
    version=get_version(package),
    description='Evaluating the optimal number of clusters for KMeans clustering using the gap statistic',
    long_description=readme,
    author=get_author(package),
    author_email=get_email(package),
    url='https://github.com/thisisandreeeee/ezcluster',
    download_url='https://github.com/thisisandreeeee/ezcluster/tarball/' + get_version(package),
    packages=[
        'ezcluster'
    ],
    package_dir={'ezcluster': 'ezcluster'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT",
    zip_safe=False,
    keywords=['kmeans', 'clustering', 'gap statistic', 'unsupervised learning', 'machine learning'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements
)
